/*
 * UNARJ.C, UNARJ, R JUNG, 02/17/93
 * Main Extractor routine
 * Copyright (c) 1991-93 by Robert K Jung.  All rights reserved.
 * modified 12 Nov 94 Thomas Seeling <ths@texbox.lahn.de>
 */

#include "unarj.h"
#include "dirutil.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/* Global variables */

UCRC    crc;
FILE   *arcfile;
FILE   *outfile;
ushort  bitbuf;
long    compsize;
long    origsize;
uchar   subbitbuf;
uchar   header[HEADERSIZE_MAX];
char    arc_name[FNAME_MAX];
char    out_dir[FNAME_MAX];
char  **exfiles;
int     overwrite;
int     command;
int     bitcount;
int     file_type;
int     no_output;
int     error_count;
int     out_pipe;
int     quiet;
int     showpath;

/* Messages */

static char *M_USAGE  [] =
{
  "Usage:  UNARJ [switches] [command] archive[.arj] [file(s)]\n",
  "               -dDEST-DIR extract to that directory\n",
  "               -o overwrite existing files\n",
  "               -p show path of files in list, -pq show only full names\n",
  "               -q quiet operation, -qq very quiet\n",
  "\n",
  "        UNARJ [l] archive      (list archive)\n",
  "        UNARJ e archive        (extract archive)\n",
  "        UNARJ t archive        (test archive)\n",
  "        UNARJ p archive        (print file in archive to stdout)\n",
  "        UNARJ x archive        (extract with pathnames)\n",
  "\n",
  "This program is copyrighted by\n",
  "\n",
  "ARJ Software                    Internet address :  robjung@world.std.com\n",
  "Robert and Susan Jung           CompuServe userid:  72077,445\n",
  "2606 Village Road West\n",
  "Norwood, Massachusetts 02062\n",
  "USA\n",
  NULL
};

char M_VERSION [] = "UNARJ 2.41.ts Copyright (c) 1991-93 Robert K Jung\n"
"modified by Thomas Seeling (ths@texbox.lahn.de) for OS/2 2.x\n\n";

char M_ARCDATE [] = "Archive created: %s";
char M_ARCDATEM[] = ", modified: %s";
char M_BADCOMND[] = "Bad UNARJ command: %s";
char M_BADCOMNT[] = "Invalid comment header";
char M_BADHEADR[] = "Bad header";
char M_BADTABLE[] = "Bad Huffman code";
char M_CANTOPEN[] = "Can't open %s";
char M_CANTREAD[] = "Can't read file or unexpected end of file";
char M_CANTWRIT[] = "Can't write file. Disk full?";
char M_CRCERROR[] = "CRC error!\n";
char M_CRCOK   [] = "CRC OK\n";
char M_DIFFHOST[] = "  Binary file!";
char M_ENCRYPT [] = "File is password encrypted, ";
char M_ERRORCNT[] = "%sFound %5d error(s)!";
char M_EXTRACT [] = "Extracting %-25s";
char M_FEXISTS [] = "%-25s exists, ";
char M_HEADRCRC[] = "Header CRC error!";
char M_NBRFILES[] = "%5d file(s)\n";
char M_NOMEMORY[] = "Out of memory";
char M_NOTARJ  [] = "%s is not an ARJ archive";
char M_PROCARC [] = "Processing archive: %s\n";
char M_SKIPPED [] = "Skipped %s\n";
char M_SUFFIX  [] = ARJ_SUFFIX;
char M_TESTING [] = "Testing    %-25s";
char M_UNKNMETH[] = "Unsupported method: %d, ";
char M_UNKNTYPE[] = "Unsupported file type: %d, ";
char M_UNKNVERS[] = "Unsupported version: %d, ";

#define get_crc()       get_longword()
#define fget_crc(f)     fget_longword(f)

#define setup_get(PTR)  (get_ptr = (PTR))
#define get_byte()      ((uchar)(*get_ptr++ & 0xff))

#define BUFFERSIZE      4096

#define ASCII_MASK      0x7F

#define CRCPOLY         0xEDB88320L

#define UPDATE_CRC(r,c) r=crctable[((uchar)(r)^(uchar)(c))&0xff]^(r>>CHAR_BIT)

/* extern functions */

int match(char *pat, char *str);

/* Local functions */

static void  make_crctable(void);
static void  crc_buf(char *str, int len);
static void  strparity(uchar *p);
static FILE  *fopen_msg(char *name, char *mode);
static int   fget_byte(FILE *f);
static uint  fget_word(FILE *f);
static ulong fget_longword(FILE *f);
static void  fread_crc(uchar *p, int n, FILE *f);
static void  decode_path(char *name);
static void  get_date_str(char *str, ulong tstamp);
static int   parse_path(char *pathname, char *path, char *entry);
static void  strncopy(char *to, char *from, int len);
static uint  get_word(void);
static ulong get_longword(void);
static long  find_header(FILE *fd);
static int   read_header(int first, FILE *fd, char *name);
static void  skip(void);
static void  unstore(void);
static int   check_flags(void);
static int   extract(void);
static int   test(void);
static uint  ratio(long a, long b);
static void  list_start(void);
static void  list_arc(int count);
static void  execute_cmd(void);
static void  help(void);

/* Local variables */

static char   filename[FNAME_MAX];
static char   comment[COMMENT_MAX];
static char   *hdr_filename;
static char   *hdr_comment;

static ushort headersize;
static uchar  first_hdr_size;
static uchar  arj_nbr;
static uchar  arj_x_nbr;
static uchar  host_os;
static uchar  arj_flags;
static short  method;
static uint   file_mode;
static ulong  time_stamp;
static short  entry_pos;
static ushort host_data;
static uchar  *get_ptr;
static UCRC   file_crc;
static UCRC   header_crc;

static long   first_hdr_pos;
static long   torigsize;
static long   tcompsize;

static int    clock_inx;

static char   *writemode[2]  = { "wb",  "w" };

static UCRC   crctable[UCHAR_MAX + 1];

/* Functions */

static void
make_crctable()
{
  uint i, j;
  UCRC r;

  for (i = 0; i <= UCHAR_MAX; i++)
    {
      r = i;
      for (j = CHAR_BIT; j > 0; j--)
        {
	  if (r & 1)
	    r = (r >> 1) ^ CRCPOLY;
	  else
	    r >>= 1;
        }
      crctable[i] = r;
    }
}

static void
crc_buf(str, len)
     char *str;
     int  len;
{
  while (len--)
    UPDATE_CRC(crc, *str++);
}

void
disp_clock()
{
  static const char clock_str[4] = { '|', '/', '-', '\\' };

  if (quiet)
    return;

  printf("(%c)\b\b\b", clock_str[clock_inx]);
  clock_inx = (clock_inx + 1) & 0x03;
}

void
error(fmt, arg)
     char *fmt;
     char *arg;
{
  putc('\n', stderr);
  fprintf(stderr, fmt, arg, error_count);
  putc('\n', stderr);
  exit(EXIT_FAILURE);
}

static void
strparity(p)
     uchar *p;
{
  while (*p)
    {
      FIX_PARITY(*p);
      p++;
    }
}

static FILE *
fopen_msg(name, mode)
     char *name;
     char *mode;
{
  FILE *fd;

  fd = file_open(name, mode);
  if (fd == NULL)
    error(M_CANTOPEN, name);
  return fd;
}

static int
fget_byte(f)
     FILE *f;
{
  int c;

  if ((c = getc(f)) == EOF)
    error(M_CANTREAD, "");
  return c & 0xFF;
}

static uint
fget_word(f)
     FILE *f;
{
  uint b0, b1;

  b0 = fget_byte(f);
  b1 = fget_byte(f);
  return (b1 << 8) + b0;
}

static ulong
fget_longword(f)
     FILE *f;
{
  ulong b0, b1, b2, b3;

  b0 = fget_byte(f);
  b1 = fget_byte(f);
  b2 = fget_byte(f);
  b3 = fget_byte(f);
  return (b3 << 24) + (b2 << 16) + (b1 << 8) + b0;
}

static void
fread_crc(p, n, f)
     uchar *p;
     int   n;
     FILE  *f;
{
  n = file_read((char *)p, 1, n, f);
  origsize += n;
  crc_buf((char *)p, n);
}

void
fwrite_txt_crc(p, n)
     uchar *p;
     int   n;
{
  uchar c;

  crc_buf((char *)p, n);
  if (no_output)
    return;

  if (file_type == TEXT_TYPE)
    {
      while (n--)
        {
	  c = *p++;
	  if (host_os != OS)
            {
	      FIX_PARITY(c);
            }
	  if (putc((int) c, outfile) == EOF)
	    error(M_CANTWRIT, "");
        }
    }
  else
    {
      if (file_write((char *)p, 1, n, outfile) != n)
	error(M_CANTWRIT, "");
    }
}

void
init_getbits()
{
  bitbuf = 0;
  subbitbuf = 0;
  bitcount = 0;
  fillbuf(2 * CHAR_BIT);
}

void
fillbuf(n)                /* Shift bitbuf n bits left, read n bits */
     int n;
{
  bitbuf = (bitbuf << n) & 0xFFFF; /* lose the first n bits */
  while (n > bitcount)
    {
      bitbuf |= subbitbuf << (n -= bitcount);
      if (compsize != 0)
        {
	  compsize--;
	  subbitbuf = (uchar) getc(arcfile);
        }
      else
	subbitbuf = 0;
      bitcount = CHAR_BIT;
    }
  bitbuf |= subbitbuf >> (bitcount -= n);
}

ushort
getbits(n)
     int n;
{
  ushort x;

  x = bitbuf >> (2 * CHAR_BIT - n);
  fillbuf(n);
  return x;
}

static void
decode_path(name)
     char *name;
{
  for ( ; *name; name++)
    {
      if (*name == ARJ_PATH_CHAR)
	*name = PATH_CHAR;
    }
}

static void
get_date_str(str, tstamp)
     char  *str;
     ulong tstamp;
{
  sprintf(str, "%04u-%02u-%02u %02u:%02u:%02u",
	  ts_year(tstamp), ts_month(tstamp), ts_day(tstamp),
	  ts_hour(tstamp), ts_min(tstamp), ts_sec(tstamp));
}

static int
parse_path(pathname, path, entry)
     char *pathname;
     char *path;
     char *entry;
{
  char *cptr, *ptr, *fptr;
  short pos;

  fptr = NULL;
  for (cptr = PATH_SEPARATORS; *cptr; cptr++)
    {
      if ((ptr = strrchr(pathname, *cptr)) != NULL &&
	  (fptr == NULL || ptr > fptr))
	fptr = ptr;
    }
  if (fptr == NULL)
    pos = 0;
  else
    pos = fptr + 1 - pathname;
  if (path != NULL)
    {
      strncpy(path, pathname, pos);
      path[pos] = NULL_CHAR;
    }
  if (entry != NULL)
    strcpy(entry, &pathname[pos]);
  return pos;
}

int matchfilename(void) {
  char **names;
  int    found=0;

  if (!exfiles)
    return 1; /* match all files */

  for (names=exfiles; *names && !found; ++names) {
    if (command=='X' || (command=='P' && showpath))
      found=match(*names,filename);    
    else
      found=match(_getname(*names),&filename[entry_pos]);
  }

  return found;
}

static void
strncopy(to, from, len)
     char *to;
     char *from;
     int  len;
{
  int i;

  for (i = 1; i < len && *from; i++)
    *to++ = *from++;
  *to = NULL_CHAR;
}

void
strlower(s)
     char *s;
{
  while (*s)
    {
      *s = (char) tolower(*s);
      s++;
    }
}

void
strupper(s)
     char *s;
{
  while (*s)
    {
      *s = (char) toupper(*s);
      s++;
    }
}

voidp *
malloc_msg(size)
     int size;
{
  char *p;

  if ((p = (char *)xmalloc(size)) == NULL)
    error(M_NOMEMORY, "");
  return (voidp *)p;
}

static uint
get_word()
{
  uint b0, b1;

  b0 = get_byte();
  b1 = get_byte();
  return (b1 << 8) + b0;
}

static ulong
get_longword()
{
  ulong b0, b1, b2, b3;

  b0 = get_byte();
  b1 = get_byte();
  b2 = get_byte();
  b3 = get_byte();
  return (b3 << 24) + (b2 << 16) + (b1 << 8) + b0;
}

static long
find_header(fd)
     FILE *fd;
{
  long arcpos, lastpos;
  int c;

  arcpos = file_tell(fd);
  file_seek(fd, 0L, SEEK_END);
  lastpos = file_tell(fd) - 2;
  if (lastpos > MAXSFX)
    lastpos = MAXSFX;
  for ( ; arcpos < lastpos; arcpos++)
    {
      file_seek(fd, arcpos, SEEK_SET);
      c = fget_byte(fd);
      while (arcpos < lastpos)
        {
	  if (c != HEADER_ID_LO) /* low order first */
	    c = fget_byte(fd);
	  else if ((c = fget_byte(fd)) == HEADER_ID_HI)
	    break;
	  arcpos++;
        }
      if (arcpos >= lastpos)
	break;
      if ((headersize = fget_word(fd)) <= HEADERSIZE_MAX)
        {
	  crc = CRC_MASK;
	  fread_crc(header, (int) headersize, fd);
	  if ((crc ^ CRC_MASK) == fget_crc(fd))
            {
	      file_seek(fd, arcpos, SEEK_SET);
	      return arcpos;
            }
        }
    }
  return -1;			/* could not find a valid header */
}

static int
read_header(first, fd, name)
     int  first;
     FILE *fd;
     char *name;
{
  ushort extheadersize, header_id;

  header_id = fget_word(fd);
  if (header_id != HEADER_ID)
    {
      if (first)
	error(M_NOTARJ, name);
      else
	error(M_BADHEADR, "");
    }

  headersize = fget_word(fd);
  if (headersize == 0)
    return 0;			/* end of archive */
  if (headersize > HEADERSIZE_MAX)
    error(M_BADHEADR, "");

  crc = CRC_MASK;
  fread_crc(header, (int) headersize, fd);
  header_crc = fget_crc(fd);
  if ((crc ^ CRC_MASK) != header_crc)
    error(M_HEADRCRC, "");

  setup_get(header);
  first_hdr_size = get_byte();
  arj_nbr = get_byte();
  arj_x_nbr = get_byte();
  host_os = get_byte();
  arj_flags = get_byte();
  method = get_byte();
  file_type = get_byte();
  (void)get_byte();
  time_stamp = get_longword();
  compsize = get_longword();
  origsize = get_longword();
  file_crc = get_crc();
  entry_pos = get_word();
  file_mode = get_word();
  host_data = get_word();

  hdr_filename = (char *)&header[first_hdr_size];
  strncopy(filename, hdr_filename, sizeof(filename));
  if (host_os != OS)
    strparity((uchar *)filename);
  if ((arj_flags & PATHSYM_FLAG) != 0)
    decode_path(filename);

  hdr_comment = (char *)&header[first_hdr_size + strlen(hdr_filename) + 1];
  strncopy(comment, hdr_comment, sizeof(comment));
  if (host_os != OS)
    strparity((uchar *)comment);

  /* if extheadersize == 0 then no CRC */
  /* otherwise read extheader data and read 4 bytes for CRC */

  while ((extheadersize = fget_word(fd)) != 0)
    file_seek(fd, (long) (extheadersize + 4), SEEK_CUR);

  return 1;			/* success */
}

static void
skip()
{
  file_seek(arcfile, compsize, SEEK_CUR);
}

static void
unstore()
{
  int n;
  long pos;
  char *buffer;

  buffer = (char *)malloc_msg(BUFFERSIZE);
  pos = file_tell(arcfile);
  disp_clock();
  n = (int)(BUFFERSIZE - (pos % BUFFERSIZE));
  n = compsize > (long)n ? n : (int)compsize;
  while (compsize > 0)
    {
      if (file_read(buffer, 1, n, arcfile) != n)
	error(M_CANTREAD, "");
      disp_clock();
      compsize -= n;
      fwrite_txt_crc((uchar *)buffer, n);
      n = compsize > BUFFERSIZE ? BUFFERSIZE : (int)compsize;
    }
  free(buffer);
}

static int
check_flags()
{
  if (arj_x_nbr > ARJ_X_VERSION)
    {
      fprintf(stderr,M_UNKNVERS, arj_x_nbr);
      fprintf(stderr,M_SKIPPED, filename);
      skip();
      return -1;
    }
  if ((arj_flags & GARBLE_FLAG) != 0)
    {
      fprintf(stderr,M_ENCRYPT);
      fprintf(stderr,M_SKIPPED, filename);
      skip();
      return -1;
    }
  if (method < 0 || method > MAXMETHOD || (method == 4 && arj_nbr == 1))
    {
      fprintf(stderr,M_UNKNMETH, method);
      fprintf(stderr,M_SKIPPED, filename);
      skip();
      return -1;
    }
  if (file_type != BINARY_TYPE && file_type != TEXT_TYPE)
    {
      fprintf(stderr,M_UNKNTYPE, file_type);
      fprintf(stderr,M_SKIPPED, filename);
      skip();
      return -1;
    }
  return 0;
}

static int
extract()
{
  static char name[FNAME_MAX];
  static char dir[FNAME_MAX];
  char *slash;

  if (check_flags())
    {
      error_count++;
      return 0;
    }

  name[0]='\0';
  if (out_dir[0])
    strcat(name,out_dir);

  if (command == 'E')
    strcat(name, &filename[entry_pos]);
  else
    {
/*      strcat(name, DEFAULT_DIR); */
      strcat(name, filename);
    }

  if (host_os != OS)
    default_case_path(name);

  strcpy(dir,name);
  slash=strrchr(dir,'\\');
  if (slash)
    *slash='\0';
  else {
    slash=strrchr(dir,'/');
    if (slash)
      *slash='\0';
  }

  if (slash && strlen(dir))
    mmkdir(dir);

  if (file_exists(name) && !overwrite)
    {
      if (quiet<2) {
	fprintf(stderr,M_FEXISTS, name);
	fprintf(stderr,M_SKIPPED, name);
      }
      skip();
      error_count++;
      return 0;
    }
  if (out_pipe)
    outfile = stdout;
  else
    outfile = file_open(name, writemode[file_type & 1]);
  if (outfile == NULL)
    {
      fprintf(stderr,M_CANTOPEN, name);
      putc('\n',stderr);
      skip();
      error_count++;
      return 0;
    }
  if (quiet<2) {
    printf(M_EXTRACT, name);
    if (host_os != OS && file_type == BINARY_TYPE)
      printf(M_DIFFHOST);
    printf("  ");
  }

  crc = CRC_MASK;

  if (method == 0)
    unstore();
  else if (method == 1 || method == 2 || method == 3)
    decode();
  else if (method == 4)
    decode_f();

  if (!out_pipe) {
    fclose(outfile);
    set_ftime_mode(name, time_stamp, file_mode, (uint) host_os);
  }

  if ((crc ^ CRC_MASK) == file_crc) {
    if (quiet<2)
      printf(M_CRCOK);
  }
  else
    {
      printf(M_CRCERROR);
      error_count++;
    }
  return 1;
}

static int
test()
{
  if (check_flags())
    return 0;

  if (quiet<2) {
    printf(M_TESTING, filename);
    printf("  ");
  }

  crc = CRC_MASK;

  if (method == 0)
    unstore();
  else if (method == 1 || method == 2 || method == 3)
    decode();
  else if (method == 4)
    decode_f();

  if ((crc ^ CRC_MASK) == file_crc)
    printf(M_CRCOK);
  else
    {
      printf(M_CRCERROR);
      error_count++;
    }
  return 1;
}

static uint
ratio(a, b)
     long a, b;
{
  int i;

  for (i = 0; i < 3; i++)
    if (a <= LONG_MAX / 10)
      a *= 10;
    else
      b /= 10;
  if ((long) (a + (b >> 1)) < a)
    {
      a >>= 1;
      b >>= 1;
    }
  if (b == 0)
    return 0;
  return (uint) ((a + (b >> 1)) / b);
}

static void
list_start()
{
  if (quiet)
    return;

  printf("Filename       Original Compressed Ratio DateTime modified CRC-32   AttrBTPMGVX\n");
  printf("------------ ---------- ---------- ----- ----------------- -------- -----------\n");
}

static void
list_arc(count)
     int count;
{
  uint r;
  int garble_mode, path_mode, volume_mode, extfil_mode, ftype, bckf_mode;
  char date_str[20], fmode_str[10];
  static char mode[5] = { 'B', 'T', '?', 'D', 'V' };
  static char pthf[2] = { ' ', '+' };
  static char pwdf[2] = { ' ', 'G' }; /* plain, encrypted */
  static char volf[2] = { ' ', 'V' };
  static char extf[2] = { ' ', 'X' };
  static char bckf[2] = { ' ', '*' };

  if (count == 0)
    list_start();

  garble_mode = ((arj_flags & GARBLE_FLAG) != 0);
  volume_mode = ((arj_flags & VOLUME_FLAG) != 0);
  extfil_mode = ((arj_flags & EXTFILE_FLAG) != 0);
  bckf_mode   = ((arj_flags & BACKUP_FLAG) != 0);
  path_mode   = (entry_pos > 0);
  r = ratio(compsize, origsize);
  torigsize += origsize;
  tcompsize += compsize;
  ftype = file_type;
  if (ftype != BINARY_TYPE && ftype != TEXT_TYPE && ftype != DIR_TYPE &&
      ftype != LABEL_TYPE)
    ftype = DIR_TYPE;
  get_date_str(date_str, time_stamp);
  strcpy(fmode_str, "    ");
  if (host_os == OS)
    get_mode_str(fmode_str, (uint) file_mode);
  if (showpath) {
    if (entry_pos) {
      if (quiet<2) {
	filename[entry_pos-1]='\0';
	printf("%s%s\\\n%-12s ",out_dir,filename,&filename[entry_pos]);
	filename[entry_pos-1]='\\';
      }
      else
	printf("%s%s",out_dir,filename);
    }
    else {
      if (out_dir[0])
	fputs(out_dir,stdout);
      if (quiet<2)
	putchar('\n');
      printf("%-12s ",filename);
    }
  }
  else
    if (strlen(&filename[entry_pos]) > 12)
      printf("%s\n             ", &filename[entry_pos]);
    else
      printf("%-12s ", &filename[entry_pos]);
  if (quiet<2)
    printf("%10ld %10ld %u.%03u %s %08lX %4s%c%c%c%u%c%c%c\n",
	   origsize, compsize, r / 1000, r % 1000, &date_str[2], file_crc,
	   fmode_str, bckf[bckf_mode], mode[ftype], pthf[path_mode], method,
	   pwdf[garble_mode], volf[volume_mode], extf[extfil_mode]);
  else
    putchar('\n');
}

static void
execute_cmd()
{
  int file_count;
  char **upper;
  char date_str[22];
  uint r;

/* uppercase all filenames from the command line */
  if (exfiles)
    for (upper=exfiles; *upper; ++upper)
      strupper(*upper);

  first_hdr_pos = 0;
  time_stamp = 0;
  first_hdr_size = FIRST_HDR_SIZE;

  arcfile = fopen_msg(arc_name, "rb");

  if (!quiet)
    printf(M_PROCARC, arc_name);

  first_hdr_pos = find_header(arcfile);
  if (first_hdr_pos < 0)
    error(M_NOTARJ, arc_name);
  file_seek(arcfile, first_hdr_pos, SEEK_SET);
  if (!read_header(1, arcfile, arc_name))
    error(M_BADCOMNT, "");
  get_date_str(date_str, time_stamp);
  if (quiet<2) {
    printf(M_ARCDATE, date_str);
    if (arj_nbr >= ARJ_M_VERSION)
      {
	get_date_str(date_str, (ulong) compsize);
	printf(M_ARCDATEM, date_str);
      }
    printf("\n");
  }

  file_count = 0;
  while (read_header(0, arcfile, arc_name)) {
    if (matchfilename())
      switch (command) {
      case 'E':
      case 'X':
	out_pipe=FALSE;
	no_output=FALSE;
	if (extract())
	  file_count++;
	else
	  skip();
	break;
      case 'L':
	list_arc(file_count++);
	skip();
	break;
      case 'T':
	no_output=TRUE;
	if (test())
	  file_count++;
	break;
      case 'P':
	out_pipe=TRUE;
	if (extract())
	  file_count++;
	break;
      }
    else
      skip();
  }

  if (command == 'L') {
    if (!quiet)
      printf("------------ ---------- ---------- ----- -----------------\n");
    if (quiet<2) {
      r = ratio(tcompsize, torigsize);
      printf("% 6d file%c %10ld %10ld %u.%03u %s\n",
	     file_count,
	     file_count==1 ? ' ' : 's',
	     torigsize, tcompsize, r / 1000,
	     r % 1000, &date_str[2]);
    }
  }
  else
    if (quiet<2)
      printf(M_NBRFILES, file_count);

  fclose(arcfile);
}

static void
help()
{
  int i;

  for (i = 0; M_USAGE[i] != NULL; i++)
    fprintf(stderr,M_USAGE[i]);
}

int
main(argc, argv)
     int  argc;
     char *argv[];
{
  int i, j, lastc;
  char *arc_p;

  out_dir[0]='\0';
  quiet=0;
  overwrite=FALSE;
  showpath=FALSE;
  out_pipe=FALSE;

  while ((i = getopt (argc, argv, "d:D:oOpPqQ")) != EOF)
    switch (i) {
    case 'd':
    case 'D':
      if (optarg && *optarg) {
	int l=strlen(optarg);

	strcpy(out_dir,optarg);
	if (out_dir[l-1]!='\\') {
	  out_dir[l++]='\\';
	  out_dir[l]='\0';
	}
      }
      break;
    case 'o':
    case 'O':
      overwrite=TRUE;
      break;
    case 'q':
    case 'Q':
      ++quiet;
      break;
    case 'p':
    case 'P':
      showpath=TRUE;
      break;
    default:
    }

  if (!quiet)
    fprintf(stderr,M_VERSION);

  if (argc == optind)
    {
      help();
      return EXIT_SUCCESS;
    }

  if (argc == optind+1)
    {
      command = 'L';
      arc_p = argv[optind];
    }
  else
    {
      if (strlen(argv[optind]) > 1)
	error(M_BADCOMND, argv[optind]);
      command = toupper(*argv[optind]);
      if (strchr("ELTXP", command) == NULL)
	error(M_BADCOMND, argv[optind]);
      arc_p = argv[++optind];
    }
/*
  else
    {
      help();
      return EXIT_FAILURE;
    }
*/

  if (++optind<argc)
    exfiles=&argv[optind];
  else
    exfiles=NULL;

  strncopy(arc_name, arc_p, FNAME_MAX);
  case_path(arc_name);
  i = strlen(arc_name);
  j = parse_path(arc_name, (char *)NULL, (char *)NULL);
  lastc = arc_name[i - 1];
  if (lastc == ARJ_DOT)
    arc_name[i - 1] = NULL_CHAR;
  else if (strchr(&arc_name[j], ARJ_DOT) == NULL)
    strcat(arc_name, M_SUFFIX);

  make_crctable();

  error_count = 0;
  clock_inx = 0;
  arcfile = NULL;
  outfile = NULL;

  execute_cmd();

  if (!quiet && error_count > 0)
    error(M_ERRORCNT, "");

  return EXIT_SUCCESS;
}

/* end UNARJ.C */
